/**
 * Copyright (c) IBM Japan 2013. All rights reserved.
 */
package net.dependableos.dcase.toolchain.rhapsodyextension.commands;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.UUID;

import javax.xml.bind.annotation.XmlAttribute;

import net.dependableos.dcase.toolchain.rhapsodyextension.DCasePlugin;
import net.dependableos.dcase.toolchain.rhapsodyextension.commands.SysML_ImportDCase.EdgeData;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.DiagramObjectFactory;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.model.ModelObjectFactory;

import com.telelogic.rhapsody.core.IRPApplication;
import com.telelogic.rhapsody.core.IRPCollection;
import com.telelogic.rhapsody.core.IRPDependency;
import com.telelogic.rhapsody.core.IRPDiagram;
import com.telelogic.rhapsody.core.IRPGraphElement;
import com.telelogic.rhapsody.core.IRPGraphNode;
import com.telelogic.rhapsody.core.IRPHyperLink;
import com.telelogic.rhapsody.core.IRPModelElement;
import com.telelogic.rhapsody.core.IRPPackage;
import com.telelogic.rhapsody.core.IRPProject;
import com.telelogic.rhapsody.core.IRPStereotype;
import com.telelogic.rhapsody.core.IRPTag;
import com.telelogic.rhapsody.core.RhapsodyAppServer;

/**
 * RhapsodyD-Case Editor̃ft@Cƃ_CAOt@CD-CasefvfGNX|[g
 */
public class SysML_ExportDCase {
	
	/** Rhapsody Application */
	protected IRPApplication app = null;
	
	/** Rhapsody Project */
	protected IRPProject prj = null;
	
	/**
	 * L[ : fvf̃XeI^Cv
	 * l  : dcase_model t@C rootBasicNode Gg xsi:type   <br>
	 *      iex. dcase:Goaljprefixiex. Goalj
	 */
	protected static Map<String, String> stereotype2XsiType = null;
	
	/**
	 * RXgN^
	 * @param app Rhapsody Application
	 */
	public SysML_ExportDCase(IRPApplication app) {
		this.app = app;
		this.prj = app.activeProject();
		
		// lazy initialization
		if (stereotype2XsiType == null) {
			stereotype2XsiType = new HashMap<String, String>();
			setStereotype2XsiType();
		}
	}
	
	/**
	 * nbV}bv stereotype2XsiType ݒ肷
	 */
	private static void setStereotype2XsiType() {
		for(int i = 0; i < SysML_ImportDCase.XSITYPE_STEREOTYPE.length; i++) {
			String xsiType = SysML_ImportDCase.XSITYPE_STEREOTYPE[i][0];
			String stereotype = SysML_ImportDCase.XSITYPE_STEREOTYPE[i][1];
			stereotype2XsiType.put(stereotype, xsiType);
		}
	}
	
	/**
	 * ӂIDԂ
	 * @return ӂID
	 */
	private String getUUID() {
		String id = UUID.randomUUID().toString();
		id = id.replace('-', '_');
		return id;
	}
	
	/**
	 * w肵^O̒l擾
	 * @param elm Ώۂ̃fGg
	 * @param tagName Ώۂ̃^O̖O
	 * @return ^O̒li^Ot^ĂȂ null Ԃj
	 */
	private static String getTag(IRPModelElement elm, String tagName){
		String tagStr = null;
		IRPTag tag = elm.getTag(tagName);
		if ( tag != null ) {
			tagStr = tag.getValue();
		}
		return tagStr;
	}
	
	/**
	 * w肵^O̒lݒ肷
	 * @param elm Ώۂ̃fGg
	 * @param tagName Ώۂ̃^O̖O
	 * @param tagValue Ώۂ̃^O̒l
	 */
	private static void setTag(IRPModelElement elm, String tagName, String tagValue){
		// ^Ol null ̏ꍇ̓^OǉȂ
		if (tagValue == null) {
			return;
		}
		
		// ̃^Oɑ݂ꍇ́A^OlXV
		IRPTag tag = (IRPTag)elm.findNestedElement(tagName, "Tag");
		if(tag == null) {
			tag = (IRPTag)elm.addNewAggr("Tag", tagName);
		}
		elm.setTagValue(tag, tagValue);
	}
	
	/**
	 * w肵fGgw肵̃^OێĂꍇAo͂ <br>
	 * ^Ot^ĂȂꍇ͋󕶎o͂
	 * @param fout o͐ PrintStream
	 * @param mSrc Ώۂ̃fGg
	 * @param attr Ώۂ̑
	 */
	private void printObligateAttr(PrintStream fout, IRPModelElement mSrc, String attr) {
		String value = getTag(mSrc, attr);
		if (value == null) {
			value = "";
		}
		fout.print("\" " + attr + "=\"");
		fout.print(normalizeXML(value));
	}
	
	/**
	 * w肵fGgw肵̃^OێĂꍇAo͂ <br>
	 * ^Ot^ĂȂꍇ͏o͂Ȃ
	 * @param fout o͐ PrintStream
	 * @param mSrc Ώۂ̃fGg
	 * @param attr Ώۂ̑
	 */
	private void printOptionalAttr(PrintStream fout, IRPModelElement mSrc, String attr) {
		String value = getTag(mSrc, attr);
		if (value != null) {
			fout.print("\" " + attr + "=\"");
			fout.print(normalizeXML(value));
		}
	}
	
	/**
	 * w肵fGgnCp[NێĂꍇÂ̈o͂ <br>
	 * nCp[Nt^ĂȂꍇ͏o͂Ȃ
	 * @param fout o͐ PrintStream
	 * @param mSrc Ώۂ̃fGg
	 * @param attr Ώۂ̑
	 */
	private void printOptionalAttrAttachment(PrintStream fout, IRPModelElement mSrc, String attr) {
		String value = getAHyperLink(mSrc);
		if (value != null) {
			fout.print("\" " + attr + "=\"");
			fout.print(normalizeXML(value));
		}
	}
	
	/**
	 * RhapsodyD-Case Editor̃ft@Cƃ_CAOt@CD-CasefvfGNX|[g
	 */
	public void exportDCase(){
		// ExportΏۃt@CI
		DCaseSelectForm form = new DCaseSelectForm();
		File choosedFile = form.saveFile();
		if (choosedFile == null) {
			return;
		}
		String choosedFileName = choosedFile.getName();
		String choosedAbsName = choosedFile.getAbsolutePath();
		
		// ex. "ACC_safety_dev"
		String modelFileName = null;
		// ex. "C:\\Documents and Settings\\Administrator\\workspace\\DCase_PlugIn\\src\\sample\\"
		String modelDirName = choosedAbsName.substring(0, choosedAbsName.length() - choosedFileName.length());
		
		final String FILE_MODEL_SUFFIX = ".dcase_model";
		final String FILE_DIAGRAM_SUFFIX = ".dcase_diagram";
		
		if(choosedFileName.endsWith(FILE_MODEL_SUFFIX)){
			modelFileName = choosedFileName.substring(0, choosedFileName.length()-FILE_MODEL_SUFFIX.length());
		}
		else if(choosedFileName.endsWith(FILE_DIAGRAM_SUFFIX)){
			modelFileName = choosedFileName.substring(0, choosedFileName.length()-FILE_DIAGRAM_SUFFIX.length());
		}
		else {
			modelFileName = choosedFileName;
		}
		
		// 蓮ŒǉD-Casevfɑ΂IDȂǂt^
		boolean success = setModelTags();
		if (!success) {
			app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "f̓ǂݍ݂Ɏs܂B\n");
			return;
		}
		
		// ft@Cɏ
		String modelID = getUUID();
		success = exportModel (modelFileName, modelDirName, modelID);
		if (!success) {
			app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "ft@C݂̏Ɏs܂B\n");
			return;
		}
		
		// _CAOt@Cɏ
		success = exportDiagram (modelFileName, modelDirName, modelID);
		if (!success) {
			app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "_CAOt@C݂̏Ɏs܂B\n");
			return;
		}
		
		app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "Export܂B\n");
	}
	
	/**
	 * ^Ot^ĂȂD-Casevfɑ΂āA^Ot^
	 * @return ^O̒ǉǂ
	 */
	private boolean setModelTags() {
		boolean ret = false;
		
		// RgnestedŒTāA^OȂ̂T
		
		// }̏ENbNƂ
		IRPDiagram d = app.getDiagramOfSelectedElement();
		if(d == null) {
			// fuEU[̏ENbNƂ
			d = (IRPDiagram) app.getSelectedElement();
		}
		IRPCollection col = d.getGraphicalElements();
		if ( col != null ){
			
			// D-Casem[h̏ꍇ
			Iterator<?> itr = col.toList().iterator();
			while ( itr.hasNext() ) {
				IRPGraphElement gSrc = (IRPGraphElement) itr.next();
				IRPModelElement mSrc = gSrc.getModelObject();
				if (mSrc == null) {
					continue;
				}
				
				// D-Casem[hInterfaceIRPCommentł
				String interfaceName = mSrc.getInterfaceName();
				if(interfaceName != null && interfaceName.equals("IRPComment")){
					
					// XeI^Cv̎擾
					IRPStereotype t = mSrc.getStereotype();
					if(t == null) {
						continue;
					}
					String stereotype = t.getDisplayName();
					
					// Ƀ^Ot^Ă΁Ã^O擾
					String type = getTag(mSrc, "type");
					String id = getTag(mSrc, "id");
					
					// ^Ot^ĂȂ΁A^Ot^
					if(type == null){
						if(stereotype.equals("Goal")) {
							type = "dcase:Goal";
						}
						else if(stereotype.equals("Evidence")) {
							type = "dcase:Evidence";
						}
						else if(stereotype.equals("Strategy")) {
							type = "dcase:Strategy";
						}
						else if(stereotype.equals("Context")) {
							type = "dcase:Context";
						}
						else if(stereotype.equals("Justification")) {
							type = "dcase:Justification";
						}
						else if(stereotype.equals("Assumption")) {
							type = "dcase:Userdef004";
						}
						else if(stereotype.equals("Undeveloped")) {
							type = "dcase:Undeveloped";
						}
						else if(stereotype.equals("Module_")) {
							type = "dcase:Userdef005";
						}
						else if(stereotype.equals("Contract")) {
							type = "dcase:Userdef006";
						}
						else if(stereotype.equals("Monitor")) {
							type = "dcase:Monitor";
						}
						else if(stereotype.equals("System")) {
							type = "dcase:System";
						}
						else if(stereotype.equals("Policy")) {
							type = "dcase:Policy";
						}
						else if(stereotype.equals("Userdef001")) {
							type = "dcase:Userdef001";
						}
						else if(stereotype.equals("Userdef002")) {
							type = "dcase:Userdef002";
						}
						else if(stereotype.equals("Userdef003")) {
							type = "dcase:Userdef003";
						}
						setTag(mSrc, "type", type);
					}
					if(id == null) {
						id = getUUID();
						setTag(mSrc, "id", id);
					}
				}
			}
			
			
			// D-CaseGbW̏ꍇ
			itr = col.toList().iterator();
			while ( itr.hasNext() ) {
				IRPGraphElement gSrc = (IRPGraphElement) itr.next();
				IRPModelElement mSrc = gSrc.getModelObject();
				if (mSrc == null) {
					continue;
				}
				
				// D-CaseGbWInterfaceIRPDependencył
				String interfaceName = mSrc.getInterfaceName();
				if(interfaceName != null && interfaceName.equals("IRPDependency")){
					
					// XeI^Cv̎擾
					IRPStereotype t = mSrc.getStereotype();
					if(t == null) {
						continue;
					}
					String stereotype = t.getDisplayName();
					
					// IRPDependencyIuWFNg̎擾
					IRPDependency dep = null;
					if( mSrc instanceof IRPDependency ) {
						dep = (IRPDependency) mSrc;
					}
					else {
						continue;
					}
					
					// Ƀ^Ot^Ă΁Ã^O擾
					String type = getTag(mSrc, "type");
					String id = getTag(mSrc, "id");
					String source = getTag(mSrc, "source");
					String target = getTag(mSrc, "target");
					
					// ^Ot^ĂȂ΁A^Ot^
					if(type == null){
						if(stereotype.equals("SupportedBy")) {
							type = "dcase:DcaseLink001";
						}
						else if(stereotype.equals("InContextOf")) {
							type = "dcase:DcaseLink002";
						}
						else if(stereotype.equals("Link1")) {
							type = "dcase:DcaseLink003";
						}
						else if(stereotype.equals("Link2")) {
							type = "dcase:DcaseLink004";
						}
						setTag(mSrc, "type", type);
					}
					if(id == null) {
						id = getUUID();
						setTag(mSrc, "id", id);
					}
					if(source == null || target == null) {
						source = "#" + getTag(dep.getDependent(), "id");
						target = "#" + getTag(dep.getDependsOn(), "id");
						setTag(mSrc, "source", source);
						setTag(mSrc, "target", target);
					}
				}
			}
		}
		
		ret = true;
		return ret;
	}
	
	/**
	 * ft@C݂̏s
	 * @param modelFileName Ώۂ̃ft@C̃t@Cigqj
	 * @param modelDirName Ώۂ̃ft@C̃tH_
	 * @param modelID Ώۂ̃ft@CID
	 * @return ݂ǂ
	 */
	private boolean exportModel(String modelFileName, String modelDirName, String modelID) {
		boolean ret = false;
		
		String fileName = modelDirName + modelFileName + ".dcase_model";
		boolean appends = false;
		PrintStream fout = null;
		boolean autoFlush = true;
		String encoding = "UTF-8";
		try{
			FileOutputStream fos = new FileOutputStream(fileName, appends);
			fout = new PrintStream(fos, autoFlush, encoding);
		}
		catch(Exception e){
			e.printStackTrace();
			System.out.println("exportModel : file IO error.");
			return ret;
		}
		
		fout.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		fout.println("<dcase:Argument xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:dcase=\"http://www.dependable-os.net/2010/03/dcase/\" id=\"" + modelID + "\">");
		
		
		// }̏ENbNƂ
		IRPDiagram d = app.getDiagramOfSelectedElement();
		if(d == null) {
			// fuEU[̏ENbNƂ
			d = (IRPDiagram) app.getSelectedElement();
		}
		IRPCollection col = d.getGraphicalElements();
		if ( col != null ){
			Iterator<?> itr = col.toList().iterator();
			while ( itr.hasNext() ) {
				IRPGraphElement gSrc = (IRPGraphElement) itr.next();
				IRPModelElement mSrc = gSrc.getModelObject();
				if (mSrc == null) {
					continue;
				}
				
				// XeI^Cv̎擾
				IRPStereotype t = mSrc.getStereotype();
				if (t== null) {
					continue;
				}
				String stereotype = t.getDisplayName();
				
				// type̎擾
				String type = getTag(mSrc, "type");
				String desc = mSrc.getDisplayName();
				
				
				if( type == null ) {
					// do nothing
					System.out.println("exportModel : unknown D-Case type: " + mSrc.getDisplayName() + ", " + stereotype);
				}
				
				// Link̏ꍇ
				else if(type.startsWith("dcase:DcaseLink00")) {
					
					fout.print("<rootBasicLink xsi:type=\"");
					fout.print(getXsiType(stereotype));
					
					printObligateAttr(fout, mSrc, "id");
					printObligateAttr(fout, mSrc, "source");
					printObligateAttr(fout, mSrc, "target");
					printObligateAttr(fout, mSrc, "name");
					
					printOptionalAttr(fout, mSrc, "attachment");
					printOptionalAttr(fout, mSrc, "status");
					printOptionalAttr(fout, mSrc, "userdef001");
					printOptionalAttr(fout, mSrc, "userdef002");
					printOptionalAttr(fout, mSrc, "userdef003");
					printOptionalAttr(fout, mSrc, "userdef004");
					printOptionalAttr(fout, mSrc, "userdef005");
					printOptionalAttr(fout, mSrc, "userdef006");
					printOptionalAttr(fout, mSrc, "userdef007");
					printOptionalAttr(fout, mSrc, "userdef008");
					printOptionalAttr(fout, mSrc, "userdef009");
					printOptionalAttr(fout, mSrc, "userdef010");
					printOptionalAttr(fout, mSrc, "userdef011");
					printOptionalAttr(fout, mSrc, "userdef012");
					printOptionalAttr(fout, mSrc, "userdef013");
					printOptionalAttr(fout, mSrc, "userdef014");
					printOptionalAttr(fout, mSrc, "userdef015");
					printOptionalAttr(fout, mSrc, "userdef016");
					
					fout.println("\"/>");
				}
				
				// Nodȅꍇ
				// iattachment̓^Oł͂ȂnCp[Nɓo^Ăj
				else {
					
					fout.print("<rootBasicNode xsi:type=\"");
					fout.print(getXsiType(stereotype));
					
					printObligateAttr(fout, mSrc, "id");
					printObligateAttr(fout, mSrc, "name");
					
					fout.print("\" desc=\"");
					fout.print(normalizeXML(desc));
					
					printOptionalAttrAttachment(fout, mSrc, "attachment");
					printOptionalAttr(fout, mSrc, "status");
					printOptionalAttr(fout, mSrc, "userdef001");
					printOptionalAttr(fout, mSrc, "userdef002");
					printOptionalAttr(fout, mSrc, "userdef003");
					printOptionalAttr(fout, mSrc, "userdef004");
					printOptionalAttr(fout, mSrc, "userdef005");
					printOptionalAttr(fout, mSrc, "userdef006");
					printOptionalAttr(fout, mSrc, "userdef007");
					printOptionalAttr(fout, mSrc, "userdef008");
					printOptionalAttr(fout, mSrc, "userdef009");
					printOptionalAttr(fout, mSrc, "userdef010");
					printOptionalAttr(fout, mSrc, "userdef011");
					printOptionalAttr(fout, mSrc, "userdef012");
					printOptionalAttr(fout, mSrc, "userdef013");
					printOptionalAttr(fout, mSrc, "userdef014");
					printOptionalAttr(fout, mSrc, "userdef015");
					printOptionalAttr(fout, mSrc, "userdef016");
					printOptionalAttr(fout, mSrc, "score");
					printOptionalAttr(fout, mSrc, "weight");
					printOptionalAttr(fout, mSrc, "nodeLink");
					printOptionalAttr(fout, mSrc, "isNormal");
					printOptionalAttr(fout, mSrc, "stakeholder");
					printOptionalAttr(fout, mSrc, "riskAnalysis");
					
					fout.println("\"/>");
				}
				
			}
		}
		else {
			System.out.println("createLink : No D-Case element is selected.");
		}
		
		fout.println("</dcase:Argument>");
		fout.close();
		
		ret = true;
		return ret;
	}
	
	/**
	 * _CAOt@C݂̏s
	 * @param diagramFileName Ώۂ̃_CAOt@C̃t@Cigqj
	 * @param diagramDirName Ώۂ̃_CAOt@C̃tH_
	 * @param modelID Ώۂ̃ft@CID
	 * @return ݂ǂ
	 */
	private boolean exportDiagram(String diagramFileName, String diagramDirName, String modelID) {
		boolean ret = false;
		
		String fileName = diagramDirName + diagramFileName + ".dcase_diagram";
		boolean appends = false;
		PrintStream fout = null;
		boolean autoFlush = true;
		String encoding = "UTF-8";
		try{
			FileOutputStream fos = new FileOutputStream(fileName, appends);
			fout = new PrintStream(fos, autoFlush, encoding);
		}
		catch(Exception e){
			e.printStackTrace();
			System.out.println("exportDiagram : file IO error.");
			return ret;
		}
		
		String diagramID = getUUID();
		String styleID = getUUID();
		
		fout.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		fout.println("<notation:Diagram xmi:version=\"2.0\" xmlns:xmi=\"http://www.omg.org/XMI\" xmlns:dcase=\"http://www.dependable-os.net/2010/03/dcase/\" xmlns:notation=\"http://www.eclipse.org/gmf/runtime/1.0.2/notation\" xmi:id=\"" + diagramID + "\" type=\"Dcase\" measurementUnit=\"Pixel\">");
		fout.println("  <styles xmi:type=\"notation:DiagramStyle\" xmi:id=\"" + styleID + "\"/>");
		fout.println("  <element xmi:type=\"dcase:Argument\" href=\"" + diagramFileName + ".dcase_model#" + modelID + "\"/>");
		fout.println("</notation:Diagram>");
		
		fout.close();
		
		ret = true;
		return ret;
	}
	
	/**
	 * dcase_model t@C rootBasicNode Gg xsi:type Ԃ
	 * @param stereotype fvf̃XeI^Cv 
	 * @return dcase_model t@C rootBasicNode Gg xsi:type 
	 */
	private static String getXsiType (String stereotype) {
		// ex. xsi:type="dcase:Goal"
		String type;
		String xsiType = stereotype;
		
		// QName  prefix iex. "dcase"jƃRt^
		type = "dcase:";
		
		if(stereotype2XsiType.containsKey(stereotype)) {
			xsiType = type + stereotype2XsiType.get(stereotype);
		}
		else {
			System.out.println("getXsiType : unknown stereotype : " + stereotype);
		}
		
		return xsiType;
	}
	
	/**
	 * w肵XMLlƂĕ\\ȕɐK
	 * @param str Ώۂ̕
	 * @return K
	 */
	private static String normalizeXML(String str) {
		String escapedStr = str;
		escapedStr = escapedStr.replaceAll("\\&", "&amp;");
		escapedStr = escapedStr.replaceAll("\\r", "&#xD;");
		escapedStr = escapedStr.replaceAll("\\n", "&#xA;");
		escapedStr = escapedStr.replaceAll(" ", "&#x20;");
		escapedStr = escapedStr.replaceAll("\t", "&#x20;");
		escapedStr = escapedStr.replaceAll("<", "&lt;");
		escapedStr = escapedStr.replaceAll(">", "&gt;");
		escapedStr = escapedStr.replaceAll("\\\"", "&quot;");
		escapedStr = escapedStr.replaceAll("\\'", "&apos;");
		return escapedStr;
	}
	
	/**
	 * w肵fGgɑ΂ĒǉnCp[N̕Ԃ <br>
	 * ԂnCp[N͍ŏɌURLSysMLfvfł <br>
	 * nCp[NȂꍇ́AnullԂ
	 * @param element Ώۂ̃fGg
	 * @return ǉnCp[N̕
	 */
	public static String getAHyperLink(IRPModelElement element) {
		String aHyperLink = null;
		IRPCollection col = element.getHyperLinks();
		if ( col != null ){
			Iterator<?> itr = col.toList().iterator();
			while ( itr.hasNext() ) {
				IRPHyperLink link = (IRPHyperLink) itr.next();
				String trgName = null;
				
				// URLݒ肵Ăꍇ
				trgName = link.getURL();
				if ( trgName != null && !trgName.equals("") ) {
					aHyperLink = trgName;
					return aHyperLink;
				}
				
				// ^[Qbgݒ肵Ăꍇ
				IRPModelElement elm = link.getTarget();
				if ( elm != null ) {
					trgName = elm.getDisplayName();
					if ( !trgName.equals("") ) {
						aHyperLink = trgName;
						return aHyperLink;
					}
				}
			}
		}
		return aHyperLink;
	}
	/**
	 * eXg
	 */
	public static void main(String[] args) {
		String str = "abc\"d\" e\tfg\r\nhi&jk'l'mn";
		System.out.println(str);
		System.out.println(normalizeXML(str));
		
		IRPApplication app = null;
		try{
			app = RhapsodyAppServer.getActiveRhapsodyApplication();
		}
		catch(Exception e){
			System.out.println("Rhapsody is not active.");
			e.printStackTrace();
			System.exit(0);
		}
		
		
		SysML_ExportDCase exporter = new SysML_ExportDCase(app);
		exporter.exportDCase();
		System.out.println("main ends.");
	}

}
