/**
 * Copyright (c) IBM Japan 2013. All rights reserved.
 */
package net.dependableos.dcase.toolchain.rhapsodyextension.commands;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import com.telelogic.rhapsody.core.IRPApplication;
import com.telelogic.rhapsody.core.IRPGraphNode;
import com.telelogic.rhapsody.core.IRPModelElement;
import com.telelogic.rhapsody.core.IRPPackage;
import com.telelogic.rhapsody.core.IRPProject;
import com.telelogic.rhapsody.core.RhapsodyAppServer;

import net.dependableos.dcase.toolchain.rhapsodyextension.DCasePlugin;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.Children;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.Diagram;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.DiagramObjectFactory;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.Edges;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.Element;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.LayoutConstraint;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.SimpleDiagram;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.diagram.SimpleDiagramObjectFactory;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.model.Argument;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.model.ModelObjectFactory;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.model.Argument.RootBasicLink;
import net.dependableos.dcase.toolchain.rhapsodyextension.jaxb.model.Argument.RootBasicNode;
import net.dependableos.dcase.toolchain.rhapsodyextension.model.DCaseDiagram;
import net.dependableos.dcase.toolchain.rhapsodyextension.model.DEdge;
import net.dependableos.dcase.toolchain.rhapsodyextension.model.DNode;

/**
 * D-Case Editor̃ft@Cƃ_CAOt@CRhapsodyD-CasefvfC|[g
 */
public class SysML_ImportDCase {
	
	/** D-Case EditorxWɑ΂RhapsodyxW̔{ */
	public static final double SCALE_FACTOR_X = 1.0;
	
	/** D-Case EditoryWɑ΂RhapsodyyW̔{ */
	public static final double SCALE_FACTOR_Y = 1.5;
	
	/** xsi:type  stereotype Ƃ̊֌W */
	static final String[][] XSITYPE_STEREOTYPE = {
		{"Goal", "Goal"},
		{"Evidence", "Evidence"},
		{"Strategy", "Strategy"},
		{"Context", "Context"},
		{"Justification", "Justification"},
		{"Userdef004", "Assumption"},
		{"Undeveloped", "Undeveloped"},
		{"Userdef005", "Module_"},
		{"Userdef006", "Contract"},
		{"Monitor", "Monitor"},
		{"System", "System"},
		{"Policy", "Policy"},
		{"Userdef001", "Userdef001"},
		{"Userdef002", "Userdef002"},
		{"Userdef003", "Userdef003"},
		{"DcaseLink001", "SupportedBy"},
		{"DcaseLink002", "InContextOf"},
		{"DcaseLink003", "Link1"},
		{"DcaseLink004", "Link2"},
	};
	
	/** dcase_model Gg̃t@Ng */
	protected static ModelObjectFactory modelFactory = null;
	
	/** dcase_diagram Gg̃t@Ng */
	protected static DiagramObjectFactory diagramFactory = null;
	
	/** Rhapsody Application */
	protected IRPApplication app = null;
	
	/** Rhapsody Project */
	protected IRPProject prj = null;
	
	/** Rhapsody Package */
	protected IRPPackage pkg = null;
	
	/**
	 * L[ : dcase_model t@C rootBasicNode Gg xsi:type 
	 *      iex. dcase:Goaljprefixiex. Goalj <br>
	 * l : fvf̃XeI^Cv
	 */
	protected static Map<String, String> xsiType2Stereotype = null;
	
	/**
	 * L[ :  "#" t^m[hID <br>
	 * l : D-Casem[h̃fGg
	 */
	protected Map<String, DNode> id2node = null;
	
	/**
	 * L[ :  "#" t^m[hID <br>
	 * l : D-Casem[h̃Otm[h
	 */
	protected Map<String, IRPGraphNode> id2graphnode = null;
	
	/**
	 * L[ :  "#" t^GbWID <br>
	 * l : D-CaseGbW֘Ãf[^
	 */
	protected Map<String, EdgeData> id2edge = null;
	
	/** GbW֘Ãf[^NX */
	protected class EdgeData {
		/** D-CaseGbW̃fGg */
		DEdge dedge;
		/**  "#" t^AGbW̌̃m[hID */
		String idSrc;
		/**  "#" t^AGbW̐̃m[hID */
		String idTrg;
		/** RXgN^ */
		EdgeData() {
			dedge = null;
			idSrc = null;
			idTrg = null;
		}
	}
	
	
	/**
	 * RXgN^
	 * @param app Rhapsody Application
	 */
	public SysML_ImportDCase(IRPApplication app) {
		this.app = app;
		this.prj = app.activeProject();
		
		// pbP[W擾
		// pbP[WIԂŃC|[gJn
		IRPModelElement selectElm = app.getSelectedElement();
		if(selectElm.getInterfaceName().equals("IRPPackage")){
			this.pkg = (IRPPackage) selectElm;
		}
		else {
			// pbP[WIĂȂƂ́Aǉ̃pbP[WDefaultƂ
			this.pkg = (IRPPackage) prj.findNestedElement("Default", "Package");
			if(this.pkg == null){
				this.pkg = prj.addPackage("Default");
			}
		}
		
		this.id2node = new HashMap<String, DNode>();
		this.id2edge = new HashMap<String, EdgeData>();
		this.id2graphnode = new HashMap<String, IRPGraphNode>();
		
		// lazy initialization
		if (xsiType2Stereotype == null) {
			xsiType2Stereotype = new HashMap<String, String>();
			setXsiType2Stereotype();
			
			modelFactory = new ModelObjectFactory();
			diagramFactory = new DiagramObjectFactory();
		}
	}
	
	/**
	 * nbV}bv xsiType2Stereotype ݒ肷
	 */
	private static void setXsiType2Stereotype() {
		for(int i = 0; i < SysML_ImportDCase.XSITYPE_STEREOTYPE.length; i++) {
			String xsiType = SysML_ImportDCase.XSITYPE_STEREOTYPE[i][0];
			String stereotype = SysML_ImportDCase.XSITYPE_STEREOTYPE[i][1];
			xsiType2Stereotype.put(xsiType, stereotype);
		}
	}
	
	/**
	 * fvf̃XeI^CvԂ
	 * @param xsiType dcase_model t@C rootBasicNode Gg xsi:type 
	 * @return fvf̃XeI^Cv
	 */
	private static String getStereotype (String xsiType) {
		// ex. xsi:type="dcase:Goal"
		String type = xsiType;
		String stereotype = xsiType;
		
		// QName  prefix iex. "dcase"jƃR
		int prefixEnd = type.indexOf(':');
		if (prefixEnd >= 0) {
			type = type.substring(prefixEnd+1, type.length());
		}
		
		if(xsiType2Stereotype.containsKey(type)) {
			stereotype = xsiType2Stereotype.get(type);
		}
		else {
			System.out.println("getStereotype : unknown xsi:type : " + xsiType);
		}
		
		return stereotype;
	}
	
	/**
	 * D-Case Editor̃ft@Cƃ_CAOt@CRhapsodyD-CasefvfC|[g
	 */
	public void importDCase(){
		// ImportΏۃt@CI
		DCaseSelectForm form = new DCaseSelectForm();
		File choosedFile = form.chooseFile();
		if (choosedFile == null) {
			return;
		}
		String choosedFileName = choosedFile.getName();
		String choosedAbsName = choosedFile.getAbsolutePath();
		
		// ex. "ACC_safety_dev"
		String modelFileName = null;
		// ex. "C:\\Documents and Settings\\Administrator\\workspace\\DCase_PlugIn\\src\\sample\\"
		String modelDirName = choosedAbsName.substring(0, choosedAbsName.length() - choosedFileName.length());
		
		final String FILE_MODEL_SUFFIX = ".dcase_model";
		final String FILE_DIAGRAM_SUFFIX = ".dcase_diagram";
		
		if(choosedFileName.endsWith(FILE_MODEL_SUFFIX)){
			modelFileName = choosedFileName.substring(0, choosedFileName.length()-FILE_MODEL_SUFFIX.length());
		}
		else if(choosedFileName.endsWith(FILE_DIAGRAM_SUFFIX)){
			modelFileName = choosedFileName.substring(0, choosedFileName.length()-FILE_DIAGRAM_SUFFIX.length());
		}
		else {
			app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "D-CaseGfB^̃ft@CĂB");
			return;
		}
		
		boolean success = importModel (modelFileName, modelDirName);
		if (!success) {
			app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "ft@C̓ǂݍ݂Ɏs܂B");
			return;
		}
		success = importDiagram (modelFileName, modelDirName);
		if (!success) {
			app.writeToOutputWindow(DCasePlugin.LOG_TITLE , "_CAOt@C̓ǂݍ݂ƕ`Ɏs܂B");
			return;
		}
	}
	
	/**
	 * ft@C̓ǂݍ݂s
	 * @param modelFileName Ώۂ̃ft@C̃t@Cigqj
	 * @param modelDirName Ώۂ̃ft@C̃tH_
	 * @return ǂݍ݂ǂ
	 */
	private boolean importModel(String modelFileName, String modelDirName) {
		boolean ret = false;
		
		// dcase_model XMLGg
		Argument argument = modelFactory.createArgument();
		
		// t@Cǂݍ
		try {
			JAXBContext context = JAXBContext.newInstance(Argument.class);
			Unmarshaller unmarshaller = context.createUnmarshaller();
			argument = (Argument) unmarshaller.unmarshal(new File(modelDirName + modelFileName + ".dcase_model"));
		}
		catch (JAXBException e) {
			e.printStackTrace();
			System.out.println("importModel : unmarshal failed.");
			return ret;
		}
		
		final String NAME_NG_REGEX = "[!\"#$%&'\\(\\)\\-=~^|\\\\@+;*:`{}\\[\\]<>,.?/ ]";
		final String NAME_OK_REGEX = "_";
		
		// D-Casem[hC|[g
		// ex. <rootBasicNode xsi:type="dcase:Goal" id="_E015SN7ZEeKZfYxKxLeCAw" name="G_1" desc="ACC͈Sł"
		//       attachment="" status="" userdef003="" userdef005="" userdef006="" userdef007=""/>
		List<RootBasicNode> nodeList = argument.getRootBasicNode();
		for(int i = 0; i < nodeList.size(); i++) {
			RootBasicNode n = nodeList.get(i);
			String ntype = n.getType();
			String nid = n.getId();
			String nname = n.getName();
			String ndesc = n.getDesc();
			
			// OƂĎgpłȂ "_" ɒu
			String nodeName = nid;
			nodeName = nodeName.replaceAll(NAME_NG_REGEX, NAME_OK_REGEX);
			
			if (ndesc == null) {
				ndesc = " "; // Xy[X
			}
			
			// D-Casem[h̃XeI^Cv擾
			String metaClass = getStereotype(ntype);
			
			// D-Casem[hǉ
			DNode node = new DNode(nodeName, ndesc, metaClass, prj, pkg);
			
			// ^O֐ݒo^
			// idesc ̓fGg̖Oƕ\ɓo^Ă̂ŁA^Oɂ͓o^Ȃj
			// iattachment ̓nCp[Nɓo^Ă̂ŁA^Oɂ͓o^Ȃj
			node.setTag("type", ntype);
			node.setTag("id", nid);
			node.setTag("name", nname);
//			node.setTag("attachment", n.getAttachment());
			node.setTag("status", n.getStatus());
			node.setTag("userdef001", n.getUserdef001());
			node.setTag("userdef002", n.getUserdef002());
			node.setTag("userdef003", n.getUserdef003());
			node.setTag("userdef004", n.getUserdef004());
			node.setTag("userdef005", n.getUserdef005());
			node.setTag("userdef006", n.getUserdef006());
			node.setTag("userdef007", n.getUserdef007());
			node.setTag("userdef008", n.getUserdef008());
			node.setTag("userdef009", n.getUserdef009());
			node.setTag("userdef010", n.getUserdef010());
			node.setTag("userdef011", n.getUserdef011());
			node.setTag("userdef012", n.getUserdef012());
			node.setTag("userdef013", n.getUserdef013());
			node.setTag("userdef014", n.getUserdef014());
			node.setTag("userdef015", n.getUserdef015());
			node.setTag("userdef016", n.getUserdef016());
			node.setTag("score", n.getScore());
			node.setTag("weight", n.getWeight());
			node.setTag("nodeLink", n.getNodeLink());
			node.setTag("isNormal", n.getIsNormal());
			node.setTag("stakeholder", n.getStakeholder());
			node.setTag("riskAnalysis", n.getRiskAnalysis());
			
			// nCp[No^
			String trgURL = n.getAttachment();
			String linkName = "Link_" + nodeName;
			String linkDispName = "LinkTo: " + trgURL;
			node.addHyperLink(linkName, linkDispName, trgURL);
			
			// }bvD-Casem[hǉ
			String shnid = "#" + nid;
			id2node.put(shnid, node);
			
//			System.out.println("[RootBasicNode] node name:" + nodeName + ", id:"+nid+", name:"+nname+", desc:"+ndesc.substring(0, Math.min(ndesc.length(), 8))+", metaCls:"+metaClass);
		}
		
		// D-CaseGbWC|[g
		// ex. <rootBasicLink xsi:type="dcase:DcaseLink001" id="_E0_p1d7ZEeKZfYxKxLeCAw" source="#_E015SN7ZEeKZfYxKxLeCAw" target="#_E015aN7ZEeKZfYxKxLeCAw" name="LINK_24"/>
		List<RootBasicLink> linkList = argument.getRootBasicLink();
		for(int i = 0; i < linkList.size(); i++) {
			RootBasicLink l = linkList.get(i);
			String ltype = l.getType();
			String lid = l.getId();
			String lsrc = l.getSource();
			String ltrg = l.getTarget();
			String lname = l.getName();
			String ldesc = l.getDesc();
			
			// OƂĎgpłȂ "_" ɒu
			String edgeName = lid;
			edgeName = edgeName.replaceAll(NAME_NG_REGEX, NAME_OK_REGEX);
			
			if (ldesc == null) {
				ldesc = " "; // Xy[X
			}
			
			// D-CaseGbW̃XeI^Cv擾
			String metaClass = getStereotype(ltype);
			DNode nSrc = null;
			DNode nTrg = null;
			if(id2node.containsKey(lsrc)) {
				nSrc = id2node.get(lsrc);
			}
			else {
				System.out.println("importModel : source node ID " + lsrc + " is unknown.");
			}
			if(id2node.containsKey(ltrg)) {
				nTrg = id2node.get(ltrg);
				
				// ˑ̖̂\
				ldesc = "LinkTo: " + nTrg.getElement().getDisplayName();
			}
			else {
				System.out.println("importModel : target node ID " + ltrg + " is unknown.");
			}
			
			// D-CaseGbWǉ
			DEdge edge = new DEdge(edgeName, ldesc, metaClass, prj, pkg, nSrc, nTrg);
			
			// ^O֐ݒo^
			// i desc D-Case EditorŕҏWłAGNX|[gɏ߂Kv邽߁A^Oɓo^j
			edge.setTag("type", ltype);
			edge.setTag("id", lid);
			edge.setTag("source", lsrc);
			edge.setTag("target", ltrg);
			edge.setTag("name", lname);
			edge.setTag("desc", l.getDesc());
			edge.setTag("attachment", l.getAttachment());
			edge.setTag("status", l.getStatus());
			edge.setTag("userdef001", l.getUserdef001());
			edge.setTag("userdef002", l.getUserdef002());
			edge.setTag("userdef003", l.getUserdef003());
			edge.setTag("userdef004", l.getUserdef004());
			edge.setTag("userdef005", l.getUserdef005());
			edge.setTag("userdef006", l.getUserdef006());
			edge.setTag("userdef007", l.getUserdef007());
			edge.setTag("userdef008", l.getUserdef008());
			edge.setTag("userdef009", l.getUserdef009());
			edge.setTag("userdef010", l.getUserdef010());
			edge.setTag("userdef011", l.getUserdef011());
			edge.setTag("userdef012", l.getUserdef012());
			edge.setTag("userdef013", l.getUserdef013());
			edge.setTag("userdef014", l.getUserdef014());
			edge.setTag("userdef015", l.getUserdef015());
			edge.setTag("userdef016", l.getUserdef016());
			
			// }bvD-CaseGbWǉ
			EdgeData ed = new EdgeData();
			ed.dedge = edge;
			ed.idSrc = lsrc;
			ed.idTrg = ltrg;
			String shlid = "#" + lid;
			id2edge.put(shlid, ed);
		}
		
		ret = true;
		return ret;
	}
	
	/**
	 * _CAOt@C̓ǂݍ݂ƁAD-Case_CAO̕`s
	 * @param diagramFileName Ώۂ̃_CAOt@C̃t@Cigqj
	 * @param diagramDirName Ώۂ̃_CAOt@C̃tH_
	 * @return ǂݍ݂ƕ`悪ǂ
	 */
	private boolean importDiagram(String diagramFileName, String diagramDirName) {
		boolean ret = false;
		
		// dcase_diagram XMLGg
		Diagram diagram = diagramFactory.createDiagram();
		
		// t@Cǂݍ
		try {
			JAXBContext context = JAXBContext.newInstance(Diagram.class);
			Unmarshaller unmarshaller = context.createUnmarshaller();
			diagram = (Diagram) unmarshaller.unmarshal(new File(diagramDirName + diagramFileName + ".dcase_diagram"));
		}
		catch (JAXBException e) {
			e.printStackTrace();
			System.out.println("importDiagram : unmarshal failed.");
			return ret;
		}
		
		// D-Casem[h`悷
		// ex.
		//   <children xmi:type="notation:Shape" xmi:id="_E0su4d7ZEeKZfYxKxLeCAw" type="1001" fontName="MS UI Gothic" lineColor="0">
		//     <children xmi:type="notation:DecorationNode" xmi:id="_E0su4t7ZEeKZfYxKxLeCAw" type="4001"/>
		//     <children xmi:type="notation:DecorationNode" xmi:id="_E0su497ZEeKZfYxKxLeCAw" type="4002"/>
		//     <children xmi:type="notation:DecorationNode" xmi:id="_E0su5N7ZEeKZfYxKxLeCAw" type="4003"/>
		//     <children xmi:type="notation:DecorationNode" xmi:id="_E0su5d7ZEeKZfYxKxLeCAw" type="4004"/>
		//     <element xmi:type="dcase:Goal" href="ACC_safety_dev.dcase_model#_E015SN7ZEeKZfYxKxLeCAw"/>
		//     <layoutConstraint xmi:type="notation:Bounds" xmi:id="_E0su5t7ZEeKZfYxKxLeCAw" x="150" y="-300"/>
		//   </children>
		String dcaseDiagramName = diagramFileName;
		String dcaseDiagramDisplayName = diagramFileName;
		DCaseDiagram d = new DCaseDiagram(dcaseDiagramName, dcaseDiagramDisplayName, prj, pkg);
		
		List<Children> childrenList = diagram.getChildren();
		for(int i = 0; i < childrenList.size(); i++) {
			Children c = childrenList.get(i);
			Element e = (Element) c.getElement();
			LayoutConstraint lc = (LayoutConstraint) c.getLayoutConstraint();
			String href = e.getHref();
			
			// dcase_model IDƏƍ D-Casem[h̃fGg擾
			int nidStart = href.indexOf('#');
			String nid = href.substring(nidStart);
			DNode node = null;
			if(id2node.containsKey(nid)) {
				node = id2node.get(nid);
			}
			else {
//				String xmiType = e.getXmiType();
//				String metaClass = getStereotype(xmiType);
//				String nname = metaClass + nid;
//				String ndispName = metaClass + nid + "<U>";
//				node = new DNode(nname, ndispName, metaClass, prj, pkg);
				
				System.out.println("importDiagram : node (ID " + nid + ") is not found in dcase_model.");
				continue;
			}
			
			// D-Casem[h`悷
			int trgX = (int) (lc.getX().intValue() * SCALE_FACTOR_X);
			int trgY = (int) (lc.getY().intValue() * SCALE_FACTOR_Y);
			int trgWidth = 80;
			int trgHeight = 60;
			IRPGraphNode gnode = d.addGraphNode(node, trgX, trgY, trgWidth, trgHeight);
			id2graphnode.put(nid, gnode);
		}
		
		// D-CaseGbW`悷
		// ex.
		//   <edges xmi:type="notation:Connector" xmi:id="_E2e3kN7ZEeKZfYxKxLeCAw" type="3001" source="_E0su4d7ZEeKZfYxKxLeCAw" target="_E015Bt7ZEeKZfYxKxLeCAw" lineColor="0">
		//     <children xmi:type="notation:DecorationNode" xmi:id="_E2e3k97ZEeKZfYxKxLeCAw" type="4061">
		//       <layoutConstraint xmi:type="notation:Location" xmi:id="_E2e3lN7ZEeKZfYxKxLeCAw" y="40"/>
		//     </children>
		//     <styles xmi:type="notation:FontStyle" xmi:id="_E2e3kd7ZEeKZfYxKxLeCAw" fontName="MS UI Gothic"/>
		//     <element xmi:type="dcase:DcaseLink001" href="ACC_safety_dev.dcase_model#_E0_p1d7ZEeKZfYxKxLeCAw"/>
		//     <bendpoints xmi:type="notation:RelativeBendpoints" xmi:id="_E2e3kt7ZEeKZfYxKxLeCAw" points="[0, 0, 0, 0]$[0, 0, 0, 0]"/>
		//   </edges>
		List<Edges> edgesList = diagram.getEdges();
		for(int i = 0; i < edgesList.size(); i++) {
			Edges edg = edgesList.get(i);
			Element e = (Element) edg.getElement();
			String href = e.getHref();
			
			// dcase_model IDƏƍ D-CaseGbW̃fGg擾
			int eidStart = href.indexOf('#');
			String eid = href.substring(eidStart);
			DEdge edge = null;
			String idSrc = null;
			String idTrg = null;
			if(id2edge.containsKey(eid)) {
				EdgeData ed = id2edge.get(eid);
				edge = ed.dedge;
				idSrc = ed.idSrc;
				idTrg = ed.idTrg;
			}
			else {
				System.out.println("importDiagram : node (ID " + eid + ") is not found in dcase_model.");
				continue;
			}
			
			// D-CaseGbW`悷
			IRPGraphNode gSrc = id2graphnode.get(idSrc);
			IRPGraphNode gTrg = id2graphnode.get(idTrg);
			if (gSrc != null && gTrg != null) {
				d.addGraphEdge(edge, gSrc, gTrg);
			}
			else {
				System.out.println("importDiagram : source node (ID " + idSrc + ") or target node (ID " + idTrg + 
						") is not found in dcase_model.");
			}
		}
		
		// 쐬D-Case}J
		d.openDiagram();
		
		ret = true;
		return ret;
	}
	
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		IRPApplication app = null;
		try{
			app = RhapsodyAppServer.getActiveRhapsodyApplication();
		}
		catch(Exception e){
			System.out.println("Rhapsody is not active.");
			e.printStackTrace();
			System.exit(0);
		}
		
		
		SysML_ImportDCase importer = new SysML_ImportDCase(app);
		importer.importDCase();
		
//		test_write();
//		test_read();
		System.out.println ("main ends.");
	}

	static void test_read() {
		
		String fileName = "ACC_safety_dev";
		String dirName = "C:\\Documents and Settings\\Administrator\\workspace\\DCase_PlugIn\\src\\sample\\";
		
		// f
		Argument argument = modelFactory.createArgument();

		// _CAO
		DiagramObjectFactory diagamFactory = new DiagramObjectFactory();
		Diagram diagram = diagamFactory.createDiagram();
		

		// t@Cǂݍ
		try {
			// ft@C
			JAXBContext context = JAXBContext.newInstance(Argument.class);

			Unmarshaller unmarshaller = context.createUnmarshaller();
			argument = (Argument) unmarshaller.unmarshal(new File(dirName + fileName + ".dcase_model"));

			
			String id = argument.getId();
			System.out.println("[Argument] id:"+id);
			
			List<RootBasicNode> nodeList = argument.getRootBasicNode();
			for(int i = 0; i < nodeList.size(); i++) {
				RootBasicNode n = nodeList.get(i);
				String nid = n.getId();
				String nname = n.getName();
				String ndesc = n.getDesc();
				System.out.println("[RootBasicNode] id:"+nid+", name:"+nname+", desc:"+ndesc);
			}
			
			List<RootBasicLink> linkList = argument.getRootBasicLink();
			for(int i = 0; i < linkList.size(); i++) {
				RootBasicLink l = linkList.get(i);
				String lid = l.getId();
				String lname = l.getName();
				String ldesc = l.getDesc();
				String lsrc = l.getSource();
				String ltrg = l.getTarget();
				System.out.println("[RootBasicLink] id:"+lid+", name:"+lname+", desc:"+ldesc+", src:"+lsrc+", trg:"+ltrg);
			}

		} catch (JAXBException e) {
			e.printStackTrace();
		}
		
		System.out.println("\n\n-----------------------------------------\n\n");
		
		// t@Cǂݍ
		try {
			// _CAOt@C
			JAXBContext context = JAXBContext.newInstance(Diagram.class);

			Unmarshaller unmarshaller = context.createUnmarshaller();
			diagram = (Diagram) unmarshaller.unmarshal(new File(dirName + fileName + ".dcase_diagram"));
			
			String did = diagram.getId();
			String dname = diagram.getName();
			System.out.println("[Diagram] id:"+did+", name:"+dname);
			
			List<Children> childrenList = diagram.getChildren();
			for(int i = 0; i < childrenList.size(); i++) {
				Children c = childrenList.get(i);
				Element e = (Element) c.getElement();
				LayoutConstraint lc = (LayoutConstraint) c.getLayoutConstraint();
				String cid = c.getXmiId();
				String ehref = e.getHref();
				String x = lc.getX().toString();
				String y = lc.getY().toString();
				System.out.println("[Children] id:"+cid+", href:"+ehref+", x:"+x+", y:"+y);
			}
			
			
			// ǂݍ񂾓et@C֏o
			// _CAOt@C
			Marshaller marshaller2 = context.createMarshaller();
			marshaller2.marshal(diagram, new File(dirName + fileName + "_out"+".dcase_diagram"));

			
			
			
		} catch (JAXBException e) {
			e.printStackTrace();
		}
		
		System.out.println("done");
	}

	static void test_write() {
		
		String fileName = "test_out";
		String dirName = "C:\\Documents and Settings\\Administrator\\workspace\\DCase_PlugIn\\src\\sample\\";
		
		
		// f
		Argument argument = modelFactory.createArgument();
		argument.setId("_XsV7AL6zEeKqg4r8yo9fPw");
//		argument.setName("argumentName_001");
//		argument.setDesc("argumentDesc_001");

		Argument.RootBasicNode contextNode1 = createContextNode("node1");
		contextNode1.setId("_E015Sd7ZEeKZfYxKxLeCAw");
		contextNode1.setName("name node 1");
		
		Argument.RootBasicNode contextNode2 = createContextNode("node2");
		contextNode2.setId("_E015St7ZEeKZfYxKxLeCAw");
		contextNode2.setName("name node 2");
		
		Argument.RootBasicNode goalNode = createGoalNode("goal1");
		goalNode.setId("_E015SN7ZEeKZfYxKxLeCAw");
		goalNode.setName("goal 2");
		goalNode.setDesc("JAPANESE 1");
		
		Argument.RootBasicNode strategyNode = createStrategyNode("strategy1");
		strategyNode.setId("_E015UN7ZEeKZfYxKxLeCAw");
		strategyNode.setName("strategy 1");

		argument.addRootBasicNode(contextNode1);
		argument.addRootBasicNode(contextNode2);
		argument.addRootBasicNode(goalNode);
		argument.addRootBasicNode(strategyNode);

		Argument.RootBasicLink link1 = createInContextOfLink("test", goalNode, contextNode1);
		argument.addRootBasicLink(link1);

		
		
		
		
		// _CAO
		SimpleDiagramObjectFactory simpleDiagramFactory = new SimpleDiagramObjectFactory();

		SimpleDiagram diagram = simpleDiagramFactory.createDiagram();
		diagram.setId(UUID.randomUUID().toString());
		diagram.setType("Dcase");
		diagram.setMeasurementUnit("Pixel");
		
		SimpleDiagram.Styles argumentStyles = simpleDiagramFactory.createDiagramStyles();
		diagram.setVersion(2.0f);
		argumentStyles.setType("ns4:DiagramStyle");
		argumentStyles.setId("_lV45IuK-EeK3gPmmwwnCzA");
		
		SimpleDiagram.Element argumentElement = simpleDiagramFactory.createDiagramElement();
		argumentElement.setType("ns3:Argument");
		argumentElement.setHref(fileName + ".dcase_model#" + argument.getId());
		
		diagram.setStyles(argumentStyles);
		diagram.setElement(argumentElement);

		// t@C֏
		try {
			// ft@C
			JAXBContext context = JAXBContext.newInstance(Argument.class);
			Marshaller marshaller = context.createMarshaller();
			
			marshaller.marshal(argument, new File(dirName + fileName + ".dcase_model"));

			// _CAOt@C
			JAXBContext context2 = JAXBContext.newInstance(SimpleDiagram.class);
			Marshaller marshaller2 = context2.createMarshaller();

			marshaller2.marshal(diagram, new File(dirName + fileName + ".dcase_diagram"));

			System.out.println("done");
		} catch (JAXBException e) {
			e.printStackTrace();
		}
	}

	private static Argument.RootBasicLink createSupportedByLink(String name,
			Argument.RootBasicNode source, Argument.RootBasicNode target) {
		Argument.RootBasicLink link1 = modelFactory.createArgumentRootBasicLink();
		link1.setName(name);
		link1.setType("ns3:DcaseLink001");
		link1.setId(UUID.randomUUID().toString());
		link1.setSource("#" + source.getId());
		link1.setTarget("#" + target.getId());
		link1.setDesc("desc");
		return link1;
	}

	private static Argument.RootBasicLink createInContextOfLink(String name,
			Argument.RootBasicNode source, Argument.RootBasicNode target) {
		Argument.RootBasicLink link1 = modelFactory.createArgumentRootBasicLink();
		link1.setName(name);
		link1.setType("ns3:DcaseLink002");
		link1.setId(UUID.randomUUID().toString());
		link1.setSource("#" + source.getId());
		link1.setTarget("#" + target.getId());
		link1.setDesc("desc");
		return link1;
	}

	private static Argument.RootBasicNode createContextNode(String name) {
		Argument.RootBasicNode node = createNode(name);
		node.setType("ns3:Context");

		return node;
	}

	private static Argument.RootBasicNode createStrategyNode(String name) {
		Argument.RootBasicNode node = createNode(name);
		node.setType("ns3:Strategy");

		return node;
	}

	private static Argument.RootBasicNode createGoalNode(String name) {
		Argument.RootBasicNode node = createNode(name);
		node.setType("ns3:Goal");

		return node;
	}

	private static Argument.RootBasicNode createNode(String name) {
		Argument.RootBasicNode node = modelFactory.createArgumentRootBasicNode();
		node.setId(UUID.randomUUID().toString());
		node.setName(name);
		node.setDesc("desc");
		
		return node;
	}
	
}
